@@ -219,6 +219,8 @@ class CValidation
         $ip = $d->GetChildVal('ip');
         $port = $d->GetChildVal('port');
 
+        $is_v6ip = ($ip == '[ANY]') || (strpos($ip, ':') !== false);
+
         $confdata = $this->_disp->Get(DInfo::FLD_ConfData);
         $lastref = $this->_disp->GetLast(DInfo::FLD_REF);
         $nodes = $confdata->GetRootNode()->GetChildren('listener');
@@ -233,7 +235,11 @@ class CValidation
             }
 
             $nodeip = $node->GetChildVal('ip');
-            if ($ip == $nodeip || $ip == 'ANY' || $nodeip == 'ANY') {
+            $is_v6node = ($nodeip == '[ANY]') || (strpos($nodeip, ':') !== false);
+            if (($ip == $nodeip)
+                    || ($ip == '[ANY]' && $is_v6node) || ($is_v6ip && $nodeip == '[ANY]')
+                    || ($ip == 'ANY' && !$is_v6node) || (!$is_v6ip && $nodeip == 'ANY')) {
+                // ANY is IPv4, [ANY] is IPv6
                 $d->SetChildErr('port', 'This port is already in use.');
                 $isValid = -1;
                 break;
@@ -303,7 +309,7 @@ class CValidation
         }
 
         $chktype = array('uint', 'name', 'vhname', 'dbname', 'admname', 'sel', 'sel1', 'sel2',
-            'bool', 'file', 'filep', 'file0', 'file1', 'filetp', 'filevh', 'path',
+            'bool', 'file', 'filep', 'file0', 'file1', 'filetp', 'filevh', 'path', 'note',
             'uri', 'expuri', 'url', 'httpurl', 'email', 'dir', 'addr', 'wsaddr', 'parse');
 
         if (!in_array($attr->_type, $chktype)) {
@@ -386,7 +392,7 @@ class CValidation
 
     protected function chkAttr_name_val($attr, $val, &$err)
     {
-        if (preg_match("/[<>&%]/", $val)) {
+        if (preg_match("/[{}<>&%]/", $val)) {
             $err = 'invalid characters in name';
             return -1;
         }
@@ -397,6 +403,15 @@ class CValidation
         return 1;
     }
 
+    protected function chkAttr_note($attr, $node)
+    {
+        if (preg_match("/[{}<]/", $node->Get(CNode::FLD_VAL), $m)) { // avoid <script, also {} for conf format
+            $node->SetErr("character $m[0] not allowed");
+            return -1;
+        }
+        return 1;
+    }
+
     protected function chkAttr_dbname($attr, $node)
     {
         return $this->chkAttr_vhname($attr, $node);
@@ -543,6 +558,22 @@ class CValidation
         return false;
     }
 
+	protected function check_cmd_invalid_str($cmd)
+	{
+		// check if it's allowed command, do not allow ' " -c -i /dev/tcp bash sh csh tcsh ksh zsh
+		$cmd = str_replace('.', 'a', $cmd); // replace . with char before pattern check
+		$pattern = '#("|\'|;|-c|-i|/dev/tcp|\Wbash\W|\Wsh\W|\Wcsh\W|\Wtcsh\W|\Wzsh\W|\Wksh\W)#';
+
+		if (preg_match($pattern, $cmd, $m)) {
+			return $m[0];
+		}
+		$cmd = str_replace('\\', '', $cmd); // remove all escape & try again
+		if (preg_match($pattern, $cmd, $m)) {
+			return $m[0];
+		}
+		return null;
+	}
+
     public function chkAttr_file_val($attr, $val, &$err)
     {
         // apply to all
@@ -562,10 +593,15 @@ class CValidation
             $path = substr($val, 0, strrpos($val, '/'));
         } else {
             $path = $val;
-            if ($attr->_type == 'file1') {
+            if ($attr->_type == 'file1') { // file1 is command
+				$invalid_str = $this->check_cmd_invalid_str($path);
+				if ($invalid_str) {
+					$err = 'Cannot contain string ' . htmlspecialchars($invalid_str, ENT_QUOTES);
+					return -1;
+				}
                 $pos = strpos($val, ' ');
                 if ($pos > 0) {
-                    $path = substr($val, 0, $pos);
+                    $path = substr($val, 0, $pos); // check first part is valid path
                 }
             }
         }
@@ -726,27 +762,41 @@ class CValidation
         $v = $node->Get(CNode::FLD_VAL);
         if (preg_match("/^([[:alnum:]._-]+|\[[[:xdigit:]:]+\]):(\d)+$/", $v)) {
             return 1;
-        } 
-        if (stripos($v, 'UDS://') !== false) {
-            $v = substr($v, 6);
+        }
+		if ($this->isUdsAddr($v)) {
+            return 1;
+		}
+
+        $node->SetErr('invalid address: correct syntax is "IPV4|IPV6_address:port" or UDS://path or unix:path');
+        return -1;
+    }
+
+	protected function isUdsAddr($v)
+	{
+		// check UDS:// unix:
+		if (preg_match('/^(UDS:\/\/|unix:)(.+)$/i', $v, $m)) {
+            $v = $m[2];
             $supportedvar = ['$SERVER_ROOT', '$VH_NAME', '$VH_ROOT', '$DOC_ROOT'];
             $v = str_replace($supportedvar, 'VAR', $v);
             if (preg_match("/^[a-z0-9\-_\/\.]+$/i", $v)) {
                 return 1;
-            } 
-        }
-        $node->SetErr('invalid address: correct syntax is "IPV4|IPV6_address:port" or UDS://path');
-        return -1;
-    }
+            }
+		}
+		return 0;
+	}
 
     protected function chkAttr_wsaddr($attr, $node)
     {
-        if (preg_match("/^((http|https):\/\/)?([[:alnum:]._-]+|\[[[:xdigit:]:]+\])(:\d+)?$/", $node->Get(CNode::FLD_VAL))) {
+		$v = $node->Get(CNode::FLD_VAL);
+        if (preg_match("/^((http|https):\/\/)?([[:alnum:]._-]+|\[[[:xdigit:]:]+\])(:\d+)?$/", $v)) {
             return 1;
-        } else {
-            $node->SetErr('invalid address: correct syntax is "[http|https://]IPV4|IPV6_address[:port]". ');
-            return -1;
-        }
+        } 
+		if ($this->isUdsAddr($v)) {
+            return 1;
+		}
+
+        $node->SetErr('invalid address: correct syntax is "[http|https://]IPV4|IPV6_address[:port]" or Unix Domain Socket address "UDS://path or unix:path".');
+        return -1;
     }
 
     protected function chkAttr_bool($attr, $node)
@@ -773,7 +823,12 @@ class CValidation
         if (preg_match($attr->_minVal, $val)) {
             return 1;
         } else {
-            $err = "invalid format \"$val\". Syntax is {$attr->_minVal} - {$attr->_maxVal}";
+			if ($attr->_maxVal) { // has parse_help
+				$err = "invalid format \"$val\". Syntax is {$attr->_minVal} - {$attr->_maxVal}";
+			} else {
+				// when no parse_help, do not show syntax, e.g. used for not allowed value.
+				$err = "invalid value \"$val\"."; 
+			}
             return -1;
         }
     }
